/* IAVLaserFeatureProxy.cpp
 *
 * Scott Livingston <SCL@utk.edu>
 * March 2009
 *
 */


#include <cmath>
#include <iostream>

#include "IAVLaserFeatureProxy.hpp"


IAVLaserFeatureProxy::IAVLaserFeatureProxy( PlayerCc::PlayerClient *aPc, uint32_t aIndex ) : LaserProxy( aPc, aIndex )
{}


void IAVLaserFeatureProxy::Update()
{
	player_point_2d_t pp;
	bool check2legs; // This name comes from Joao's pv3d code.

	if (!this->IsValid() || !this->IsFresh())
		return;

	laserFeature.fdata.clear();
	laserFeature.ranges.clear();
	laserFeature.point_xy.clear();

	for (uint i = 0; i < this->GetCount(); i++) {
		laserFeature.ranges.push_back( this->GetRange(i) );
		pp = this->GetPoint(i);
		laserFeature.point_xy.push_back( xy( pp.px, pp.py ) );
	}

	laserFeature.Segmentation();

	for (uint i = 0; i < laserFeature.segments.size(); i++) {

		// Fit lines
		laserFeature.RecursiveLineFitting( laserFeature.segments[i].begin,
										   laserFeature.segments[i].end );

		// Fit circles
		check2legs = !laserFeature.FitArc( laserFeature.segments[i].begin,
										   laserFeature.segments[i].end );

		// Find legs
		laserFeature.FindLeg( laserFeature.segments[i].begin,
							  laserFeature.segments[i].end, check2legs );

	}

	// Leg tracking!
	legTrack.fdata.clear();
	legTrack.fdata = laserFeature.fdata;
	legTrack.LCReset();
	legTrack.CreateIncoming();
	legTrack.LegMatchInTime();
	legTrack.LegClean();
	legTrack.LegCreate();
	legTrack.PersonUpdate();
	legTrack.PersonCreate();
	legTrack.FillFeature();
}


void IAVLaserFeatureProxy::PrintFeatureCounts() const
{
	uint numCircles, numLines, numPLeg, numP2Legs; 

	// Print features found
	numCircles = numLines = numPLeg = numP2Legs = 0;
	for (uint i = 0; i < laserFeature.fdata.size(); i++) {
		switch (laserFeature.fdata[i].type) {
		case MIARN_FEATURE_TYPE_ARC:
			numCircles++;
			break;
		case MIARN_FEATURE_TYPE_LINE:
			numLines++;
			break;
		case MIARN_FEATURE_TYPE_PLEG:
			numPLeg++;
			std::cout << "PLEG found at ( " << laserFeature.fdata[i].pos[0] << ", "
					  << laserFeature.fdata[i].pos[1] << " ), at angle "
					  << atan2f( laserFeature.fdata[i].pos[1], laserFeature.fdata[i].pos[0] )*180/3.14
					  << " deg" << std::endl;
			break;
		case MIARN_FEATURE_TYPE_2LEGS:
			numP2Legs++;
			std::cout << "2LEGS found at ( " << laserFeature.fdata[i].pos[0] << ", "
					  << laserFeature.fdata[i].pos[1] << " ), at angle "
					  << atan2f( laserFeature.fdata[i].pos[1], laserFeature.fdata[i].pos[0] )*180/3.14
					  << " deg" << std::endl;
			break;
		case MIARN_FEATURE_TYPE_SEGMENT:
			// Ignore these.
			break;
		default:
			std::cerr << "Error: unrecognized feature type code: " << laserFeature.fdata[i].type << std::endl;
		}
	}
	if (numCircles > 0 || numLines > 0 || numPLeg > 0 || numP2Legs > 0)
		std::cout << "Found on this cycle"
				  << "\n  circles: " << numCircles
				  << "\n  lines: " << numLines
				  << "\n  pleg: " << numPLeg
				  << "\n  p2legs: " << numP2Legs << std::endl;

	// List extracted "people" (actually, pairs of legs)
	for (uint i = 0; i < legTrack.fdata_out.size(); i++) {
		switch (legTrack.fdata_out[i].type) {
		case MIARN_FEATURE_TYPE_LEG:
			std::cout << "Leg found at ( " << legTrack.fdata_out[i].pos[0] << ", "
					  << legTrack.fdata_out[i].pos[1] << " ), at angle "
					  << atan2f( legTrack.fdata_out[i].pos[1], legTrack.fdata_out[i].pos[0] )*180/3.14
					  << " deg" << std::endl;
			break;
		case MIARN_FEATURE_TYPE_PERSON:
			std::cout << "Person found at ( " << legTrack.fdata_out[i].pos[0] << ", "
					  << legTrack.fdata_out[i].pos[1] << " ), at angle "
					  << atan2f( legTrack.fdata_out[i].pos[1], legTrack.fdata_out[i].pos[0] )*180/3.14
					  << " deg" << std::endl;
			break;
		default:
			std::cerr << "Error: unrecognized feature type code: " << laserFeature.fdata[i].type << std::endl;
		}
	}
}


miarn_fiducial_item IAVLaserFeatureProxy::GetFeature( int index ) const
{
	miarn_fiducial_item empty; // In case given index is out of range.

	if (index >= 0 && index < laserFeature.fdata.size())
		return laserFeature.fdata[index];
	else
		return empty;
}


miarn_fiducial_item IAVLaserFeatureProxy::GetPersonFeature( int index ) const
{
	miarn_fiducial_item empty; // In case given index is out of range.

	if (index >= 0 && index < legTrack.fdata_out.size())
		return legTrack.fdata_out[index];
	else
		return empty;
}
