/* PersonTracker.cpp - Definitions for PersonTracker class
 *                     (see PersonTracker.hpp for details).
 *
 * Scott Livingston <SCL@utk.edu>
 * April-May 2009
 *
 * NOTES: - Change magic numbers to object parameters... later.
 *
 *        - When executing tracking Update method, assume that the
 *          robot has been moving with current robotTransVelocity and
 *          robotRotVelocity since the last target person position
 *          update. This may lead to (significant?) errors in the
 *          estimated transformation to the robot coordinate
 *          frame. Note that this is only relevant if the tracker
 *          state is not ``unknown'' when Update method is called.
 *
 */


#include <cmath>
#include <vector>

#include "PersonTracker.hpp"

//const double laser_dist_error_factor = 1/30.;
//const double speed_error_factor = 0.5;


PersonTracker::PersonTracker( PlayerCc::PlayerClient *aPc_laser, uint32_t aIndex_laser,
							  PlayerCc::PlayerClient *aPc_camera, uint32_t aIndex_camera,
							  const IplImage *initImg, CvRect initRegion, int numBins )
	: iavLaserProxy( aPc_laser, aIndex_laser ),
	  cameraProxy( aPc_camera, aIndex_camera ),
	  kot( initImg, initRegion, numBins, 1., false ),
	  cht( initImg, initRegion, numBins ),
	  maxPersonRadius( 1. ),
	  maxPersonSpeed( 2. ),
	  pixel2rad( 0.004 ), // (found by trial-and-error)
	  maxVisionLaserMismatch( 0.0872665 ) // ~5 deg
{
	trackedState = PERSONTRACK_UNINIT;
	robotTransVelocity = robotRotVelocity = 0.;
	prevImg = NULL;
}


PersonTracker::~PersonTracker()
{
	if (prevImg != NULL)
		cvReleaseImage( &prevImg );
}


bool PersonTracker::InitTarget( const IplImage *initImg, CvRect initRegion, int numBins )
{
	if (initImg == NULL || numBins < 1)
		return false;

	if (!kot.InitTargetModel( initImg, initRegion, numBins, 1., false ))
		return false;

	cht.InitTargetHist( initImg, initRegion, numBins );

	return true;
}


int PersonTracker::Update()
{
	unsigned int i;
	bool updateFlag; /* Indicate when to break from an updating cycle;
						specifically, when the tracker state is
						PERSONTRACK_KNOWN, this flag shows when one of
						the attempted update routines succeeded (and
						hence, this Update method can return). */
	double x, y; // (x,y) coordinate of previous estimated position (simplifies tracking calculations)

	/* Used when looking for a minimum distance (resp., angle) between
	   previously estimated target position (resp., between laser and
	   vision-based position estimates when tracker is
	   uninitialized). */
	int minIndex;
	double minDist;
	double elapsedTime;

	std::vector<int> singleLegIndices, legPairIndices;

	CvRect targetRegion;
	int cam_width, cam_height, cam_depth;

	// Buffer the leg feature index vectors to improve performance
	singleLegIndices.reserve(20);
	legPairIndices.reserve(20);

	/* If tracker state is ``known,'' then attempt to apply transform
	   to previously estimated target person position by determining
	   likely change in robot coordinate frame, given nonzero
	   translational or rotational velocities. */
	if (trackedState == PERSONTRACK_KNOWN
		&& (robotTransVelocity != 0. || robotRotVelocity != 0.)) {
		elapsedTime= GetTimeSinceUpdate(); // Find elapsed time since last update
		x = estimatedPos.rho*cos( estimatedPos.theta );
		y = estimatedPos.rho*sin( estimatedPos.theta );
		if (robotRotVelocity == 0.) { // pure forward or backward movement
			x -= elapsedTime*robotTransVelocity;
			estimatedPos.rho = sqrt( x*x + y*y );
			estimatedPos.theta = atan2( y, x );
		} else if (robotTransVelocity == 0.) { // pure rotation
			estimatedPos.theta -= elapsedTime*robotRotVelocity;
			while (estimatedPos.theta > M_PI) // Bound angle to target between -pi and pi
				estimatedPos.theta -= 2*M_PI;
			while (estimatedPos.theta < -M_PI)
				estimatedPos.theta += 2*M_PI;
		} else { // both rotation and translation (i.e., both nonzero)

			if (robotTransVelocity < 0) { // moving in reverse
				x += (robotTransVelocity/robotRotVelocity)*sin(elapsedTime*robotRotVelocity);
				y += (robotTransVelocity/robotRotVelocity)*(1-cos(elapsedTime*robotRotVelocity));
			} else { // moving forward
				x -= (robotTransVelocity/robotRotVelocity)*sin(elapsedTime*robotRotVelocity);
				y -= (robotTransVelocity/robotRotVelocity)*(1-cos(elapsedTime*robotRotVelocity));
			}
			
			estimatedPos.rho = sqrt( x*x + y*y );
			estimatedPos.theta = atan2( y, x ) - elapsedTime*robotRotVelocity;
			while (estimatedPos.theta > M_PI) // Bound angle to target between -pi and pi
				estimatedPos.theta -= 2*M_PI;
			while (estimatedPos.theta < -M_PI)
				estimatedPos.theta += 2*M_PI;

		}
	}

	// Update laser and vision-based trackers:

	cam_width = cameraProxy.GetWidth();
	cam_height = cameraProxy.GetHeight();
	cam_depth = cameraProxy.GetDepth();

	// Fail on invalid camera image width, height, or depth
	if (cam_width <= 0 || cam_height <= 0 || cam_depth <= 0)
		return -1;

	if (UpdateVision()) {
		targetRegion = kot.GetTargetRegion();
	} else {
		return -1;
	}

	iavLaserProxy.Update();

	/* Push indices of single leg features and "leg pair" features
	   (much less reliable) into vectors for convenient (and fast)
	   referencing. */
	for (i = 0; i < iavLaserProxy.GetNumFeatures(); i++) {
		if (iavLaserProxy.GetFeature(i).type == MIARN_FEATURE_TYPE_PLEG) {
			singleLegIndices.push_back( i );
		} else if (iavLaserProxy.GetFeature(i).type == MIARN_FEATURE_TYPE_2LEGS) {
			legPairIndices.push_back( i );
		}
	}

	// Attempt to update the tracked person position
	switch (trackedState) {

	case PERSONTRACK_UNINIT:
		{
			/* Attempt to find a single leg feature that aligns with target angle estimated
			   by vision-based tracker. */
			minDist = -1.0; // indicate no minimum found yet.
			for (i = 0; i < singleLegIndices.size(); i++) {
				if (fabs(atan2( iavLaserProxy.GetFeature(singleLegIndices[i]).pos[1],
								iavLaserProxy.GetFeature(singleLegIndices[i]).pos[0] )
						 - (cam_width/2 - (targetRegion.x+targetRegion.width/2))*pixel2rad) <= maxVisionLaserMismatch) {
					// A match has been found!
					if (minDist < 0
						|| fabs(atan2( iavLaserProxy.GetFeature(singleLegIndices[i]).pos[1],
									   iavLaserProxy.GetFeature(singleLegIndices[i]).pos[0] )
								- (cam_width/2 - (targetRegion.x+targetRegion.width/2))*pixel2rad) < minDist) {
						// Update minimum
						minDist = fabs(atan2( iavLaserProxy.GetFeature(singleLegIndices[i]).pos[1],
											  iavLaserProxy.GetFeature(singleLegIndices[i]).pos[0] )
									   - (cam_width/2 - (targetRegion.x+targetRegion.width/2))*pixel2rad);
						minIndex = i;
					}
					
				}
			}

			if (minDist >= 0) {
				trackedState = PERSONTRACK_KNOWN;

				// Initialize target position estimate using the extracted laser feature.
				estimatedPos.rho = sqrt( iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[1]*iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[1]
										 + iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[0]*iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[0] );
				estimatedPos.theta = atan2( iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[1],
												iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[0] );
				//estimatedPos.uncertainty = estimatedPos.rho*estimatedPos.rho*laser_dist_error_factor;
				if (gettimeofday( &timestamp, NULL ) == -1)
					return -1.;
				break;
			}

			//if (trackedState == PERSONTRACK_KNOWN)
			//	break;

			// Now try the "leg pair" features
			minDist = -1.0; /* Indicate no minimum found yet. (This
							   should have carried over from above,
							   but for clarity, set minDist to
							   negative value again here.) */
			for (i = 0; i < legPairIndices.size(); i++) {
				if (fabs(atan2( iavLaserProxy.GetFeature(legPairIndices[i]).pos[1],
								iavLaserProxy.GetFeature(legPairIndices[i]).pos[0] )
						 - (cam_width/2 - (targetRegion.x+targetRegion.width/2))*pixel2rad) <= maxVisionLaserMismatch) {
					// A match is found!
					if (minDist < 0 ||
						fabs(atan2( iavLaserProxy.GetFeature(legPairIndices[i]).pos[1],
									iavLaserProxy.GetFeature(legPairIndices[i]).pos[0] )
							 - (cam_width/2 - (targetRegion.x+targetRegion.width/2))*pixel2rad) < minDist) {
						minDist = fabs(atan2( iavLaserProxy.GetFeature(legPairIndices[i]).pos[1],
											  iavLaserProxy.GetFeature(legPairIndices[i]).pos[0] )
									   - (cam_width/2 - (targetRegion.x+targetRegion.width/2))*pixel2rad);
						minIndex = i;
					}
					
				}
			}

			if (minDist >= 0) {
				trackedState = PERSONTRACK_KNOWN;

				// Initialize target position estimate using the extracted laser feature.
				estimatedPos.rho = sqrt( iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[1]*iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[1]
										 + iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[0]*iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[0] );
				estimatedPos.theta = atan2( iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[1],
												iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[0] );
				//estimatedPos.uncertainty = estimatedPos.rho*estimatedPos.rho*laser_dist_error_factor;
				if (gettimeofday( &timestamp, NULL ) == -1)
					return -1.;
				break;
			}

		}
		break;

	case PERSONTRACK_KNOWN:
		{
			/* Just to get something working quickly, this currently
			   ONLY uses the single leg features, followed by "leg
			   pair" features for person tracking (based on previous
			   position estimate). */

			updateFlag = false;

			x = estimatedPos.rho*cos( estimatedPos.theta );
			y = estimatedPos.rho*sin( estimatedPos.theta );

			// Is a reasonable single leg feature nearby the previous estimate?
			minDist = -1.0; // Indicate no minimum distance found yet.
			for (i = 0; i < singleLegIndices.size(); i++) {
				if (sqrt( (iavLaserProxy.GetFeature(singleLegIndices[i]).pos[0] - x)
						  *(iavLaserProxy.GetFeature(singleLegIndices[i]).pos[0] - x)
						  + (iavLaserProxy.GetFeature(singleLegIndices[i]).pos[1] - y)
						  *(iavLaserProxy.GetFeature(singleLegIndices[i]).pos[1] - y) )
					< maxPersonSpeed*GetTimeSinceUpdate() + 2*maxPersonRadius) {
					// Match!
					if (minDist < 0 || 
						sqrt( (iavLaserProxy.GetFeature(singleLegIndices[i]).pos[0] - x)
							  *(iavLaserProxy.GetFeature(singleLegIndices[i]).pos[0] - x)
							  + (iavLaserProxy.GetFeature(singleLegIndices[i]).pos[1] - y)
							  *(iavLaserProxy.GetFeature(singleLegIndices[i]).pos[1] - y) ) < minDist) {
						minDist = sqrt( (iavLaserProxy.GetFeature(singleLegIndices[i]).pos[0] - x)
										*(iavLaserProxy.GetFeature(singleLegIndices[i]).pos[0] - x)
										+ (iavLaserProxy.GetFeature(singleLegIndices[i]).pos[1] - y)
										*(iavLaserProxy.GetFeature(singleLegIndices[i]).pos[1] - y) );
						minIndex = i;
					}

				}
			}
			if (minDist >= 0) {
				

				estimatedPos.rho = sqrt( iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[1]*iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[1]
										 + iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[0]*iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[0] );
				estimatedPos.theta = atan2( iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[1],
											iavLaserProxy.GetFeature(singleLegIndices[minIndex]).pos[0] );
				if (gettimeofday( &timestamp, NULL ) == -1)
					return -1.;
				updateFlag = true;
				break;
			}

			//if (updateFlag) // Did we match a single leg feature to previous position estimate?
			//	break;

			// Is a "leg pair" feature nearby?
			minDist = -1.0; // Indicate no minimum distance found yet.
			for (i = 0; i < legPairIndices.size(); i++) {
				if (sqrt( (iavLaserProxy.GetFeature(legPairIndices[i]).pos[0] - x)
						  *(iavLaserProxy.GetFeature(legPairIndices[i]).pos[0] - x)
						  + (iavLaserProxy.GetFeature(legPairIndices[i]).pos[1] - y)
						  *(iavLaserProxy.GetFeature(legPairIndices[i]).pos[1] - y) )
					< maxPersonSpeed*GetTimeSinceUpdate() + 2*maxPersonRadius) {
					// Match!
					if (minDist < 0 ||
						sqrt( (iavLaserProxy.GetFeature(legPairIndices[i]).pos[0] - x)
							  *(iavLaserProxy.GetFeature(legPairIndices[i]).pos[0] - x)
							  + (iavLaserProxy.GetFeature(legPairIndices[i]).pos[1] - y)
							  *(iavLaserProxy.GetFeature(legPairIndices[i]).pos[1] - y) ) < minDist) {
						minDist = sqrt( (iavLaserProxy.GetFeature(legPairIndices[i]).pos[0] - x)
										*(iavLaserProxy.GetFeature(legPairIndices[i]).pos[0] - x)
										+ (iavLaserProxy.GetFeature(legPairIndices[i]).pos[1] - y)
										*(iavLaserProxy.GetFeature(legPairIndices[i]).pos[1] - y) );
						minIndex = i;
					}
					
				}
			}
			if (minDist >= 0) {
				estimatedPos.rho = sqrt( iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[1]*iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[1]
										 + iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[0]*iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[0] );
				estimatedPos.theta = atan2( iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[1],
											iavLaserProxy.GetFeature(legPairIndices[minIndex]).pos[0] );
				if (gettimeofday( &timestamp, NULL ) == -1)
					return -1.;
				updateFlag = true;
				break;
			}

			//if (updateFlag)
			//	break;

			// If no reasonable leg features found, then attempt to use vision-based tracking.
			if (fabs(estimatedPos.theta
					 - (cam_width/2 - (targetRegion.x+targetRegion.width/2))*pixel2rad) <= 2*maxVisionLaserMismatch) {
				// Match! (only update theta (i.e., angle) component of position)
				estimatedPos.theta = (cam_width/2 - (targetRegion.x+targetRegion.width/2))*pixel2rad;
				if (gettimeofday( &timestamp, NULL ) == -1)
					return -1.;
				updateFlag = true;
				break;
			}

			// At this point, we have failed to find the tracked person; reset the tracker.
			trackedState = PERSONTRACK_UNINIT;
			break;
		}
		break;

	default: // Unrecognized tracker state!
		return -1;

	}

	return trackedState;
}


bool PersonTracker::CalibrateCam( PlayerCc::PlayerClient *robot,
								  double x, double y,
								  unsigned int maxStabilizationCount )
{
	double theta; // of given (x,y) point converted to polar coordinates
	float u; // horizontal position of target in image with respect to center
	unsigned int i; // counter for stabilization loop

	/* Maximum distance (in image coordinates) by which vision-based
	   target region may change between updates to be considered
	   ``stable.'' */
	const double maxDist = 10.;

	CvRect targetRegion, prevTargetRegion;

	int cam_width = cameraProxy.GetWidth();

	if (maxStabilizationCount == 0) // Ignore silly calls.
		return false;

	// Initialize candidate target region.
	robot->Read();
	if (UpdateVision()) {
		prevTargetRegion = kot.GetTargetRegion();
	} else {
		return false;
	}

	// Wait for target (candidate) region to stabilize in camera image.
	for (i = 0; i < maxStabilizationCount; i++) {
		robot->Read();
		if (UpdateVision()) {
			targetRegion = kot.GetTargetRegion();
		} else {
			continue;
		}
		if ((targetRegion.x-prevTargetRegion.x)*(targetRegion.x-prevTargetRegion.x)
			+ (targetRegion.y-prevTargetRegion.y)*(targetRegion.y-prevTargetRegion.y) <= maxDist*maxDist)
			break;
		prevTargetRegion = targetRegion;
	}
	if (i == maxStabilizationCount)
		return false; // Vision component failed to stabilize.

	theta = atan2( y, x );
	u = cam_width/2. - (targetRegion.x+targetRegion.width/2.);
	
	if (u == 0 || theta*u < 0) // `check for reasonableness'
		return false;

	pixel2rad = theta/u;

	return true;
}


bool PersonTracker::ResetTrackedPos()
{
	trackedState = PERSONTRACK_UNINIT;
	return true;
}


target_position_t PersonTracker::GetEstimate() const
{
	return estimatedPos;
}


unsigned char PersonTracker::GetTrackerState() const
{
	return trackedState;
}


double PersonTracker::GetLastUpdateTime() const
{
	return timestamp.tv_sec + timestamp.tv_usec/(float)(1e6);
}


double PersonTracker::GetTimeSinceUpdate() const
{
	timeval current_time;
	if (gettimeofday( &current_time, NULL ) == -1)
		return -1.;

	return current_time.tv_sec + current_time.tv_usec/(float)(1e6)
		- timestamp.tv_sec + timestamp.tv_usec/(float)(1e6);
}


bool PersonTracker::SetMaxPersonRadius( float newRadius )
{
	if (newRadius < 0)
		return false;

	maxPersonRadius = newRadius;
	return true;
}


bool PersonTracker::SetMaxPersonSpeed( float newSpeed )
{
	if (newSpeed < 0)
		return false;

	maxPersonSpeed = newSpeed;
	return true;
}


bool PersonTracker::SetPixels2RadRatio( float newRatio )
{
	if (newRatio < 0)
		return false;

	pixel2rad = newRatio;
	return true;
}


bool PersonTracker::SetMaxVisionLaserMismatch( float newMaxMismatch )
{
	if (newMaxMismatch < 0)
		return false;

	maxVisionLaserMismatch = newMaxMismatch;
	return true;
}


bool PersonTracker::SetVisionInitCenter( int x0, int y0 )
{
	return kot.SetInitCenter( x0, y0 );
}


int PersonTracker::GetVisionCenterX() const
{
	return kot.GetCenterX();
}


int PersonTracker::GetVisionCenterY() const
{
	return kot.GetCenterY();
}


CvRect PersonTracker::GetTargetRegion() const
{
	return kot.GetTargetRegion();
}


bool PersonTracker::SetVisionBandwidth( double newBW ) // Relevant for kernel-based (vision) object tracker
{
	return kot.SetBandwidth( newBW );
}


double PersonTracker::GetVisionBandwidth() const
{
	return kot.GetBandwidth();
}


bool PersonTracker::SetRobotVelocity( float newTransVelocity, float newRotVelocity )
{
	if (newTransVelocity < 0 || newRotVelocity < 0)
		return false;

	robotTransVelocity = newTransVelocity;
	robotRotVelocity = newRotVelocity;
	
	return true;
}


//////////////////////////////////
// Private Methods
//////////////////////////////////

bool PersonTracker::UpdateVision()
{
	IplImage *img;
	uint8_t *imgBuf;
	int cam_width, cam_height, cam_depth;
	int i, j;

	// Update vision-based tracker:

	// Pull current image from camera and pass it to vision-based trackers.
	cam_width = cameraProxy.GetWidth();
	cam_height = cameraProxy.GetHeight();
	cam_depth = cameraProxy.GetDepth();

	// Fail on invalid camera image width, height, or depth
	if (cam_width <= 0 || cam_height <= 0 || cam_depth <= 0)
		return false;

	// Image buffer and IplImage structure
	imgBuf = new uint8_t[cam_width*cam_height*cam_depth];
	img = cvCreateImage( cvSize(cam_width, cam_height), IPL_DEPTH_8U, 3 );
	if (imgBuf == NULL || img == NULL) { // failed to allocate necessary space
	    if (imgBuf != NULL)
			delete [] imgBuf;
		if (img != NULL)
			cvReleaseImage( &img );
		return false;
	}

	cameraProxy.Decompress();
	cameraProxy.GetImage( imgBuf );

	// Copy image from camera proxy
	for (i = 0; i < cam_width; i++) {
		for (j = 0; j < cam_height; j++) {
			img->imageData[cam_width * j*3 + i*3 + 0] = (char)imgBuf[cam_width * j*3 + i*3 + 2];
			img->imageData[cam_width * j*3 + i*3 + 1] = (char)imgBuf[cam_width * j*3 + i*3 + 1];
			img->imageData[cam_width * j*3 + i*3 + 2] = (char)imgBuf[cam_width * j*3 + i*3 + 0];
		}
	}

	// Handle special case of unitialized previous image.
	if (prevImg == NULL)
		prevImg = cvCloneImage( img );

	// Color histogram-based tracking
	cht.GetMotionRegions( prevImg, img );
	//cht.MaxNCandidateRegions( 7 ); // Area-based pruning
	//cht.CompareRegions( img );
	cht.CompareDerivRegions( img );
	cht.OrderByArea(); //Area-based sorting of "target regions"
	cht.PruneByArea( 1000 );
	cht.OrderByCorrelation(); //Correlation magnitude-based sorting of "target regions"

	/* If high correlation region found with color
	   histogram-based tracker, then use it to initialize the
	   kernel-based object tracker. */
	if (cht.GetCount() > 0 && cht[0].r > .8
		&& ((cht[0].rect.x + cht[0].rect.width/2 - kot.GetCenterX())*(cht[0].rect.x + cht[0].rect.width/2 - kot.GetCenterX())
			+ (cht[0].rect.y + cht[0].rect.height/2 - kot.GetCenterY())*(cht[0].rect.y + cht[0].rect.height/2 - kot.GetCenterY()) > 1000)) {
		kot.SetInitCenter( cht[0].rect.x + cht[0].rect.width/2,
						   cht[0].rect.y + cht[0].rect.height/2 );
	} else if (kot.GetCenterX() < 0 || kot.GetCenterX() > cam_width
			|| kot.GetCenterY() < 0 || kot.GetCenterY() > cam_height) {
		kot.SetInitCenter( cam_width/2, cam_height/2 );
	}

	kot.Track( img );

	// In preparation for the next update
	cvCopy( img, prevImg );

	// Free image buffers.
	delete [] imgBuf;
	if (img != NULL)
		cvReleaseImage( &img );

	return true;
}
