/* vd.cc - video dump, to facilitate logging camera data and
 *         coordinating with the built-in laser sensor data logging in
 *         Player.
 *
 * NOTES: - The output file is camera_dump-T.dat , where T is the
 *          number of seconds since the Epoch (as returned by time()
 *          function).
 *
 * Scott Livingston
 * Originally created: 13 Feb 2009
 * Current: 6 May 2009
 */


#include <cv.h>
#include <highgui.h>

#include <libplayerc++/playerc++.h>
using namespace PlayerCc;

#include <sys/time.h>
#include <time.h>
#include <stdio.h>

#include <iostream>
#include <fstream>
using namespace std;


int main( int argc, char *argv[] )
{
	uint cam_width, cam_height, cam_depth;
	uint8_t *imgBuf;
	IplImage *img;
	uint i, j; // generic counters
	time_t sec_time = time(NULL); //seconds since the Epoch!
	uint frame_count = 0;
	uint max_frames = 200; //default number of camera frames to capture

	char time_str[30];
	struct timeval start_time, current_time;

	sprintf( time_str, "camera_dump-%d.dat", (int)sec_time );
	ofstream img_logfile( time_str );

	PlayerClient robot( "localhost" );
	CameraProxy cp( &robot, 0 );
	LogProxy lp( &robot, 0 );

	// Determine desired number of camera frames to capture, if given
	if (argc == 2) {
		max_frames = atoi( argv[1] );
	} else if (argc > 2) {
		cout << "Usage: " << argv[0] << " [num-frames-to-capture]" << endl;
		return 1;
	}

	//set appropriate Player replace rule on server
	robot.SetReplaceRule( true, PLAYER_MSGTYPE_DATA, -1, -1 );

	// determine camera image specs
	robot.Read();
	cam_width = cp.GetWidth();
	cam_height = cp.GetHeight();
	cam_depth = cp.GetDepth();

	// image buffer and OpenGL image structure
	imgBuf = new uint8_t[cam_width*cam_height*cam_depth];
	img = cvCreateImage( cvSize(cam_width, cam_height), IPL_DEPTH_8U, 3 );

	// open window for displaying camera image
	//cvNamedWindow( "testWin", CV_WINDOW_AUTOSIZE );

	try {

		sprintf( time_str, "laser_dump-%d.dat", (int)sec_time );
		lp.SetFilename( time_str );

		if (gettimeofday( &start_time, NULL ) == -1) {
			perror( "main, gettimeofday" );
			return -1;
		}
		lp.SetWriteState(1); //start player sensor data logging
		
		// loop until ESCAPE key is pressed
		while (cvWaitKey(33) != 27 && frame_count < max_frames) {

			// capture most recent camera image
			robot.Read();
			if (gettimeofday( &current_time, NULL ) == -1) {
				perror( "main, gettimeofday" );
				return -1;
			}

			cp.Decompress();
			cp.GetImage( imgBuf );

			// copy image from camera proxy
			for (i = 0; i < cam_width; i++) {
				for (j = 0; j < cam_height; j++) {
					img->imageData[cam_width * j*3 + i*3 + 0] = (char)imgBuf[cam_width * j*3 + i*3 + 2];
					img->imageData[cam_width * j*3 + i*3 + 1] = (char)imgBuf[cam_width * j*3 + i*3 + 1];
					img->imageData[cam_width * j*3 + i*3 + 2] = (char)imgBuf[cam_width * j*3 + i*3 + 0];
				}
			}

			frame_count++;

			// Enter new entry in camera image logfile
			img_logfile << (current_time.tv_sec-start_time.tv_sec) + (double)(current_time.tv_usec-start_time.tv_usec)/1e6
						<< ' ' << frame_count << endl;
			
			sprintf( time_str, "camdat-%04d.jpg", frame_count );
			cvSaveImage( time_str, img );
			//cvShowImage( "testWin", img );

		}

	} catch (PlayerError e) {
		cerr << e << endl;
		return -1;
	}

	lp.SetWriteState(0);
	img_logfile.close();

	//cvDestroyWindow( "testWin" );

	// release image buffer and image structure
	cvReleaseImage( &img );
	delete [] imgBuf;

	return 0;
}
