/* ColorHistTracker.cpp
 *
 * Scott Livingston <SCL@utk.edu>
 * March 2009
 *
 */


#include <algorithm>
#include <vector>
#include <cmath>

#include <iostream>

// Only for debugging. cvShowImage is called from ColorHistTracker::GetMotionRegions.
//#include <highgui.h> 

#include "ColorHistTracker.hpp"


ColorHistTracker::ColorHistTracker( float threshold, double normalFactor )
{
	// Silently handle bad arguments

	if (threshold < 0 || threshold > 1)
		compThreshold = 0.4;
	else
		compThreshold = threshold;

	if (normalFactor < 0)
		this->normalFactor = 1024;
	else
		this->normalFactor = normalFactor;
	
	targetHist = NULL;
}

ColorHistTracker::ColorHistTracker( const IplImage *initImg, CvRect initRegion, int numBins,
									float threshold, double normalFactor )
{
	// Silently handle bad arguments
	if (threshold < 0 || threshold > 1)
		compThreshold = 0.4;
	else
		compThreshold = threshold;

	if (normalFactor < 0)
		this->normalFactor = 1024;
	else
		this->normalFactor = normalFactor;

	targetHist = NULL;
	InitTargetHist( initImg, initRegion, numBins, threshold, normalFactor );
}

ColorHistTracker::~ColorHistTracker()
{
	if (targetHist != NULL)
		cvReleaseHist( &targetHist );
}


void ColorHistTracker::InitTargetHist( const IplImage *initImg, CvRect initRegion, int numBins,
									   float threshold, double normalFactor )
{
	float hueRange[] = {0, 180};
	float *ranges[] = {hueRange};
	IplImage *hsv, *hue;

	// Ignore empty calls or bad arguments
	if (initImg == NULL
		|| numBins < 1 || threshold < 0 || threshold > 1
		|| normalFactor < 0)
		return;

	// Clear and reset prior target/histogram values
	this->normalFactor = normalFactor;
	compThreshold = threshold;
	this->numBins = numBins;

	if (targetHist != NULL)
		cvReleaseHist( &targetHist );

	objRegions.clear();
	candidateRegions.clear();
	targetHistDeriv.clear();

	// Pull hue from given image
	hsv = cvCreateImage( cvGetSize(initImg), 8, 3 );
	hue = cvCreateImage( cvGetSize(initImg), 8, 1 );
	cvCvtColor( initImg, hsv, CV_BGR2HSV );
	cvSplit( hsv, hue, 0, 0, 0 );

	// Allocate and calculate color histogram
	cvSetImageROI( hue, initRegion );
	targetHist = cvCreateHist( 1, &numBins, CV_HIST_ARRAY, ranges, 1 );
	cvCalcHist( &hue, targetHist );
	cvNormalizeHist( targetHist, normalFactor );

	// Release temporarily allocated image structures
	cvReleaseImage( &hsv );
	cvReleaseImage( &hue );

	// Find forward difference of target color histogram
	for (int i = 0; i < numBins-1; i++) {
		targetHistDeriv.push_back( cvQueryHistValue_1D( targetHist, i+1 ) - cvQueryHistValue_1D( targetHist, i ) );
	}

	// Finally, calculate and store variances
	targetVar = CalcVariance( targetHist );
	targetDerivVar = CalcDerivVariance( targetHistDeriv );
}


TrackRect ColorHistTracker::GetRegion( uint index ) const
{
	if (index > objRegions.size()-1) {
		std::cerr << "Error in ColorHistTracker[]: request for entry of region array out-of-bounds." << std::endl;
		TrackRect tmp;
		tmp.rect = cvRect( 0, 0, 0, 0 );
		tmp.r = 0;
		return tmp;
	}
	
	return objRegions[index];
}	


uint ColorHistTracker::FindMax() const
{
	if (objRegions.size() < 1)
		return -1;

	uint max_i = 0;

	for (uint i = 1; i < objRegions.size(); i++)
		if (objRegions[i].r > objRegions[max_i].r)
			max_i = i;

	return max_i;
}


void ColorHistTracker::GetMotionRegions( const IplImage *prevImg, const IplImage *img,
										 bool globalMotionCancel )
{
	IplImage *prevImgGray, *imgGray,
		*diff;
	CvMemStorage *storage;
	CvSeq *seq;

	// NOTE: for now, the globalMotionCancel parameter is ignored.
	globalMotionCancel = false;

	if (globalMotionCancel) {
		// ...
	}

	// 1. Difference of given images
	// 2. Apply 3x3 "blur" (mean) filter
	// 3. Threshold result (UNSW students suggest threshold value between 25 and 35 (out of 255).)
	// 4. Contour detection to cluster connected pixels into groups, find minimal bounding boxes,
	//    then merge "nearby" boxes to account for poor clustering results...
	// The resulting regions (or "bounding boxes") are the candidates for consideration.

	candidateRegions.clear();
	
	// Convert given images to grayscale
	prevImgGray = cvCreateImage( cvGetSize(img), 8, 1 );
	imgGray = cvCreateImage( cvGetSize(img), 8, 1 );
	diff = cvCreateImage( cvGetSize(prevImg), 8, 1 );

	cvCvtColor( prevImg, prevImgGray, CV_BGR2GRAY );
	cvCvtColor( img, imgGray, CV_BGR2GRAY );

	// Absolute difference, Gaussian smoothing, and then apply threshold
	cvAbsDiff( prevImgGray, imgGray, diff );
	cvSmooth( diff, diff );
	cvThreshold( diff, diff, 25, 255, CV_THRESH_BINARY ); // UNSW students recommend between 25 and 35.
	cvSmooth( diff, diff, CV_GAUSSIAN, 3, 3 );

	// DEBUG (visualize result of difference-smooth-threshold step)
	//cvShowImage( "prep", diff );

	storage = cvCreateMemStorage();
	cvClearMemStorage( storage );

	cvFindContours( diff, storage, &seq, sizeof(CvContour),
					CV_RETR_EXTERNAL, CV_CHAIN_APPROX_NONE, cvPoint(0,0) );

	for (; seq; seq = seq->h_next)
		candidateRegions.push_back( cvBoundingRect( seq ) );

	// Free temporarily allocated structures
	cvReleaseImage( &prevImgGray );
	cvReleaseImage( &imgGray );
	cvReleaseImage( &diff );
	cvReleaseMemStorage( &storage );
}


void ColorHistTracker::CompareRegions( const IplImage *img )
{
	std::vector<CvRect>::const_iterator candIt;
	TrackRect currentTrackReg;

	float hueRange[] = {0, 180};
	float *ranges[] = {hueRange};
	IplImage *hsv, *hue;
	CvHistogram *candHist;

	double correlation; // Corresponds to r field in TrackRect structure

	// Ignore silly arguments
	if (img == NULL || candidateRegions.size() < 1)
		return;

	// Clear previous results, in preparation for new findings
	objRegions.clear();

	// Pull hue from given image
	hsv = cvCreateImage( cvGetSize(img), 8, 3 );
	hue = cvCreateImage( cvGetSize(img), 8, 1 );
	cvCvtColor( img, hsv, CV_BGR2HSV );
	cvSplit( hsv, hue, 0, 0, 0 );

	for (candIt = candidateRegions.begin(); candIt != candidateRegions.end(); candIt++) {

		// Allocate and calculate color histogram
		cvSetImageROI( hue, *candIt );
		candHist = cvCreateHist( 1, &numBins, CV_HIST_ARRAY, ranges, 1 );
		cvCalcHist( &hue, candHist );
		cvNormalizeHist( candHist, normalFactor );

		/* Find correlation and decide whether to promote this region
		   to "tracking region" list (private member vector objRegions). */
		correlation = CalcCovariance( candHist )/sqrt( targetVar*CalcVariance( candHist ) );
		if (correlation > compThreshold) {
			currentTrackReg.rect = *candIt;
			currentTrackReg.r = correlation;
			objRegions.push_back( currentTrackReg );
		}

		// Free this histogram
		cvReleaseHist( &candHist );
	}

	// Release temporarily allocated image structures
	cvReleaseImage( &hsv );
	cvReleaseImage( &hue );
}


void ColorHistTracker::CompareDerivRegions( const IplImage *img )
{
	std::vector<CvRect>::const_iterator candIt;
	TrackRect currentTrackReg;

	float hueRange[] = {0, 180};
	float *ranges[] = {hueRange};
	IplImage *hsv, *hue;
	CvHistogram *candHist;
	std::vector<double> candHistDeriv;

	double correlation; // Corresponds to r field in TrackRect structure

	// Ignore silly arguments
	if (img == NULL || candidateRegions.size() < 1)
		return;

	// Clear previous results, in preparation for new findings
	objRegions.clear();

	// Pull hue from given image
	hsv = cvCreateImage( cvGetSize(img), 8, 3 );
	hue = cvCreateImage( cvGetSize(img), 8, 1 );
	cvCvtColor( img, hsv, CV_BGR2HSV );
	cvSplit( hsv, hue, 0, 0, 0 );

	for (candIt = candidateRegions.begin(); candIt != candidateRegions.end(); candIt++) {

		// Allocate and calculate color histogram and its derivative
		cvSetImageROI( hue, *candIt );
		candHist = cvCreateHist( 1, &numBins, CV_HIST_ARRAY, ranges, 1 );
		cvCalcHist( &hue, candHist );
		cvNormalizeHist( candHist, normalFactor );
		candHistDeriv.clear();
		for (int i = 0; i < numBins-1; i++) {
			candHistDeriv.push_back( cvQueryHistValue_1D( candHist, i+1 ) - cvQueryHistValue_1D( candHist, i ) );
		}

		/* Find correlation and decide whether to promote this region
		   to "tracking region" list (private member vector objRegions). */
		correlation = CalcDerivCovariance( candHistDeriv )/sqrt( targetDerivVar*CalcDerivVariance( candHistDeriv ) );
		if (correlation > compThreshold) {
			currentTrackReg.rect = *candIt;
			currentTrackReg.r = correlation;
			objRegions.push_back( currentTrackReg );
		}

		// Free this histogram
		cvReleaseHist( &candHist );
	}

	// Release temporarily allocated image structures
	cvReleaseImage( &hsv );
	cvReleaseImage( &hue );
}


/* Compare function used by call to sort (STL implementation of
   quicksort algorithm) in ColorHistTracker::MaxNCandidateRegions. */
bool RectAreaCompare( CvRect rect1, CvRect rect2 )
{ return (rect1.width*rect1.height) > (rect2.width*rect2.height); }

void ColorHistTracker::MaxNCandidateRegions( uint maxNum )
{
	// Handle special cases
	if (maxNum == 0) {
		candidateRegions.clear();
		return;
	} else if (maxNum >= candidateRegions.size()) {
		return; // Leave unchanged
	}

	// Sort candidate regions vector in descending area order
	sort( candidateRegions.begin(), candidateRegions.end(), RectAreaCompare );

	// ...and resize as requested
	candidateRegions.resize(maxNum);
}


bool RectAreaCompare_TrackRect( TrackRect tr1, TrackRect tr2 )
{ return (tr1.rect.width*tr1.rect.height) > (tr2.rect.width*tr2.rect.height); }

void ColorHistTracker::OrderByArea()
{ sort( objRegions.begin(), objRegions.end(), RectAreaCompare_TrackRect ); }


bool CorrCompare( TrackRect trect1, TrackRect trect2 )
{ return trect1.r*trect1.r > trect2.r*trect2.r; } // Ensure non-negativity

void ColorHistTracker::OrderByCorrelation()
{ sort( objRegions.begin(), objRegions.end(), CorrCompare ); }


void ColorHistTracker::PruneByArea( double minArea )
{
	if (minArea < 0)
		return;

	for (uint i = 0; i < objRegions.size(); i++) {
		if (objRegions[i].rect.width*objRegions[i].rect.height < minArea) {
			objRegions.erase(objRegions.begin()+i);
			i--;
		}
	}
}


////////////////////////////////////////
// Private methods:
////////////////////////////////////////

double ColorHistTracker::CalcVariance( const CvHistogram *hist )
{
	double mean = 0,
		variance = 0;

	if (hist == NULL)
		return -1; // Ignore bad arguments

	for (int i = 0; i < numBins; i++) {
		mean += cvQueryHistValue_1D( hist, i );
	}
	mean /= numBins;
	for (int i = 0; i < numBins; i++) {
		variance += (cvQueryHistValue_1D( hist, i ) - mean)*(cvQueryHistValue_1D( hist, i ) - mean);
	}
	variance /= numBins;

	return variance;
}

double ColorHistTracker::CalcDerivVariance( std::vector<double> histDeriv )
{
	double mean = 0,
		variance = 0;
	std::vector<double>::const_iterator it;

	for (it = histDeriv.begin(); it != histDeriv.end(); it++) {
		mean += *it;
	}
	mean /= histDeriv.size();
	for (it = histDeriv.begin(); it != histDeriv.end(); it++) {
		variance += (*it - mean)*(*it - mean);
	}
	variance /= histDeriv.size();

	return variance;
}

double ColorHistTracker::CalcCovariance( const CvHistogram *hist )
{
	double meanX = 0, meanY = 0,
		cov = 0;

	if (hist == NULL)
		return -1; // Ignore bad arguments

	for (int i = 0; i < numBins; i++) {
		meanX += cvQueryHistValue_1D( targetHist, i );
		meanY += cvQueryHistValue_1D( hist, i );
	}
	meanX /= numBins;
	meanY /= numBins;
	for (int i = 0; i < numBins; i++) {
		cov += (cvQueryHistValue_1D( targetHist, i ) - meanX)*(cvQueryHistValue_1D( hist, i ) - meanY);
	}
	cov /= numBins;

	return cov;
}

double ColorHistTracker::CalcDerivCovariance( std::vector<double> histDeriv )
{
	double meanX = 0, meanY = 0,
		cov = 0;
	uint histLength;

	if (histDeriv.size() != targetHistDeriv.size())
		return -1; // Ignore bad arguments

	histLength = histDeriv.size();

	for (uint i = 0; i < histLength; i++) {
		meanX += targetHistDeriv[i];
		meanY += histDeriv[i];
	}
	meanX /= histLength;
	meanY /= histLength;
	for (uint i = 0; i < histLength; i++) {
		cov += (targetHistDeriv[i] - meanX)*(histDeriv[i] - meanY);
	}
	cov /= histLength;

	return cov;
}
