/* ColorHistTracker.hpp - Class to encapsulate target tracking by
 *                        comparing color histograms of image regions.
 *
 * Scott Livingston <SCL@utk.edu>
 * March 2009
 *
 * NOTES: - Add other methods to identify candidate regions, such as
 *          based on Haar classifiers (a la OpenCV routines).
 */

#ifndef _COLORHISTTRACKER_HPP_
#define _COLORHISTTRACKER_HPP_


#include <vector>

#include <cv.h>


typedef unsigned int uint;


// Associate OpenCV rectangle with correlation (of histograms)
typedef struct {
	CvRect rect;
	double r; // Correlation with original color histogram (or its derivative)
} TrackRect;


class ColorHistTracker {

public:

	// Constructors and destructor
	ColorHistTracker( float threshold = 0.4, double normalFactor = 1024 );
	ColorHistTracker( const IplImage *initImg, CvRect initRegion, int numBins = 16,
					  float threshold = 0.4, double normalFactor = 1024 );
	~ColorHistTracker();

	// (Re)initialize target color histogram and related data structures
	void InitTargetHist( const IplImage *initImg, CvRect initRegion, int numBins = 16,
						 float threshold = 0.4, double normalFactor = 1024 );

	/* Find regions of interest (ROIs) by differencing the given
	   images to (likely poorly) detect motion. The parameter
	   globalMotionCancel indicates whether to attempt to ignore
	   global motion, also called background motion (e.g., caused by a
	   moving camera). */
	void GetMotionRegions( const IplImage *prevImg, const IplImage *img,
						   bool globalMotionCancel = true );

	/* Reduce number of candidate regions (as found, e.g., by
	   GetMotionRegions) to at most the maxNum largest, in terms of area. */
	void MaxNCandidateRegions( uint maxNum );

	/* Compare candidate regions (e.g., found using GetMotionRegions
	   method) with original target region color histogram. */
	void CompareRegions( const IplImage *img );

	/* Similar to CompareRegions method, but find correlation with
	   forward difference of color histograms. */
	void CompareDerivRegions( const IplImage *img );

	/* Place "object regions" (i.e., objRegions vector) in order of
	   decreasing rectangle area (so, objRegions[0] has the largest area). */
	void OrderByArea();

	// Same as OrderByArea, but sort by correlation.
	void OrderByCorrelation();

	/* Delete all "object regions" (also called "target regions") of
	   area less than minArea */
	void PruneByArea( double minArea );

	// Facilitate access to likely target regions
	uint GetCount() const { return this->objRegions.size(); }
	TrackRect GetRegion( uint index ) const;
	TrackRect operator[] (uint index) const { return GetRegion( index ); }

	// Convenience routines:
	uint FindMax() const; /* Return index of likely target region with
							 highest correlation.  Returns -1 on error, in
							 particular if objRegions is empty. */

private:

	// Covariance between targetHist and given histogram
	double CalcCovariance( const CvHistogram *hist ); // Assumes hist has same number of bins as targetHist
	// Cov with derivative of target color histogram
	double CalcDerivCovariance( std::vector<double> histDeriv ); 

	// Find variance of given histogram or of derivative of that histogram
	double CalcVariance( const CvHistogram *hist ); // Also assumes number of bins is numBins
	double CalcDerivVariance( std::vector<double> histDeriv );

	double normalFactor; // Normalization factor for comparing histograms
	float compThreshold; /* Threshold for correlation between
							candidate region and known (at
							initialization) target. */
	std::vector<TrackRect> objRegions; /* Likely target regions, based on
									 correlation with original given
									 sample image */
	std::vector<CvRect> candidateRegions; // Regions of interest

	CvHistogram *targetHist; // Color histogram of target
	int numBins;
	double targetVar; // Target color histogram variance

	std::vector<double> targetHistDeriv; // Derivative of the target color histogram
	double targetDerivVar; // Variance of the derivative of the target color histogram

};


#endif
