/* testCST_player.cpp - Example using CvCamShiftTracker class that
 *                      reads images from camera device a la Player.
 *
 * NOTES: - For definition of CvCamShiftTracker, see cvaux.hpp in the
 *          main include directory for opencv (probably
 *          /usr/include/opencv); it basically encapsulates typical
 *          use of the camshift algorithm as exemplified by the
 *          example file camshiftdemo.oc (distributed with the OpenCV
 *          library).
 *
 * Scott Livingston <SCL@utk.edu>
 * Originally created: 8 April 2009
 *
 */


#include <cv.h>
#include <cv.hpp>
#include <cvaux.hpp>
#include <highgui.h>

#include <libplayerc++/playerc++.h>
using namespace PlayerCc;

#include <iostream>
using namespace std;


typedef unsigned int uint;


int main( int argc, char **argv )
{
	IplImage *img,
		*dispImage; // Use bounding boxes for visual feedback
	CvRect cst_result;
	int hdims = 16;

	uint cam_width, cam_height, cam_depth;
	uint8_t *imgBuf;

	uint i, j; // generic counters

	if (argc != 6) {
		cerr << "Usage: " << argv[0] << " init_image x0 y0 width height" << endl;
		return 1;
	}

	PlayerClient robot( "localhost" );
	CameraProxy camp( &robot, 0 );

	// Determine camera image specs
	robot.Read();
	cam_width = camp.GetWidth();
	cam_height = camp.GetHeight();
	cam_depth = camp.GetDepth();

	// Attempt to load target initialization image
	if ((img = cvLoadImage( argv[1] )) == NULL) {
		cerr << "Error: Could not open image file: " << argv[1] << endl;
		return -1;
	}

	// Verify sign of given region arguments
	if (atoi(argv[2]) < 0 || atoi(argv[3]) < 0
		|| atoi(argv[4]) < 0 || atoi(argv[5]) < 0 ) {
		cerr << "Error: Given region specs (x0, y0, width, height) must be non-negative." << endl;
		return -1;
	}

	// Attempt to instantiate and initialize a CvCamShiftTracker object
	CvCamShiftTracker cst;
	if (!cst.set_window( cvRect( atoi(argv[2]), atoi(argv[3]),
								 atoi(argv[4]), atoi(argv[5]) ) )
		|| !cst.set_hist_dims( 1, &hdims )
		|| !cst.update_histogram( img )) {
		cerr << "Error: Could initialize target region." << endl;
		return -1;
	}
	cvReleaseImage( &img ); // Init target image is no longer needed.
	img = NULL;

	// Initialize HSV bounds for camshift tracker
	/*cst.set_min_ch_val( 0, 0 );
	cst.set_min_ch_val( 1, 30 );
	cst.set_min_ch_val( 2, 10 );
	cst.set_max_ch_val( 0, 180 );
	cst.set_max_ch_val( 1, 256 );
	cst.set_max_ch_val( 2, 256 ); */

	// Image buffer and IplImage structure
	imgBuf = new uint8_t[cam_width*cam_height*cam_depth];
	img = cvCreateImage( cvSize(cam_width, cam_height), IPL_DEPTH_8U, 3 );

	// Open a window for visual feedback and enter main (tracking) loop
	cvNamedWindow( "track results", 1 );
	//cvNamedWindow( "prep", 1 );
	dispImage = NULL;

	try {
		while (cvWaitKey(10) != 27) {

			robot.Read();
			camp.Decompress();
			camp.GetImage( imgBuf );

			// Copy image from camera proxy
			for (i = 0; i < cam_width; i++) {
				for (j = 0; j < cam_height; j++) {
					img->imageData[cam_width * j*3 + i*3 + 0] = (char)imgBuf[cam_width * j*3 + i*3 + 2];
					img->imageData[cam_width * j*3 + i*3 + 1] = (char)imgBuf[cam_width * j*3 + i*3 + 1];
					img->imageData[cam_width * j*3 + i*3 + 2] = (char)imgBuf[cam_width * j*3 + i*3 + 0];
				}
			}

			// camshift tracking
			if (!cst.track_object( img )) {
				cerr << "Error: CvCamShiftTracker::track_object method returned error." << endl;
				return -1;
			}
			cst_result = cst.get_window();

			// Draw rectangles on image to show found "target regions"
			if (dispImage)
				cvReleaseImage( &dispImage );
			dispImage = cvCloneImage( img );
			
			cvRectangle( dispImage,
						 cvPoint(cst_result.x, cst_result.y),
						 cvPoint(cst_result.x+cst_result.width, cst_result.y+cst_result.height),
						 CV_RGB( 255, 0, 0 ), 5 );
			cvShowImage( "track results", dispImage );
		
			// For debugging, freeze frame when at least one "target region" is found
			/*if (cht.GetCount() > 0)
			  cvWaitKey(0);
			*/

		}
	} catch (PlayerError e) {
		cerr << e << endl;
		return -1;
	}

	cvDestroyWindow( "track results" );
	//cvDestroyWindow( "prep" );
	
	// Release image buffer and structures
	if (img != NULL)
		cvReleaseImage( &img );
	delete [] imgBuf;

	return 0;
}

